Alfrin

Engineering Notebook

Alfrin Poulose

Software EngineerAISystemsPerformanceOpen Source

I build reliable, high-performance software systems. My core interest lies in understanding abstractions deeply—from the database engine up to distributed AI inference layers—and removing friction for other developers.

Currently Exploring: Distributed Consensus & Hardware Acceleration

Notebook

Technical deep-dives into computer systems, databases, and compilers.

VIEW ALL

Selected Projects

Production-ready open-source engines and system tools.

PROJECT / 01

A High Performance LLM Inference Engine

A custom inference engine designed to optimize memory bandwidth and reduce latency for local LLM deployment. Built in C++ and CUDA to achieve 90%+ theoretical memory bandwidth utilization.

LATENCY~45% REDUCTION
THROUGHPUT3,200 TOK/SEC
C++CUDAPyTorchSIMDKV CACHE
kernel_fused_attention.cuCUDA v12.2
__global__ void flash_attention_v2_kernel(
    const half* __restrict__ Q,
    const half* __restrict__ K,
    const half* __restrict__ V,
    half* __restrict__ O,
    const int seq_len,
    const int head_dim
) {
    // Shared memory allocations for tiling
    extern __shared__ half s_mem[];
    half* s_Q = s_mem;
    half* s_K = &s_mem[BLOCK_SIZE * head_dim];

    // Grid-stride coalesced load into SRAM
    int tid = threadIdx.x;
    int bid = blockIdx.x;
    ...
}
Architecture: NVIDIA Ampere/HopperZero Copy Memory

About & Approach

memory

Low-Level Mastery

Understanding hardware characteristics—L1/L2 caches, register pressure, memory bandwidth—to write code that respects physical limits.

layers

Clear Abstractions

Designing APIs and architectures that conceal operational complexity while keeping high performance accessible to upstream users.

speed

Empirical Profiling

Never guessing performance bottlenecks. Driven by micro-benchmarks, trace analysis, flamegraphs, and hardware telemetry.